#include <stdio.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "app4lcd.h"
#include "app4delay.h"


/* 
** constants/macros 
*/
#define PIN(x) (*(&x - 2))  /* address of data direction register of port x */
#define DDR(x) (*(&x - 1))  /* address of input register of port x          */


#define delay(x) delay_us(x)
#define lcd_e_delay()   __asm__ __volatile__( "rjmp 1f\n 1:" );

#define lcd_e_high()    LCD_CTL_PORT  |=  _BV(LCD_E_PIN);
#define lcd_e_low()     LCD_CTL_PORT  &= ~_BV(LCD_E_PIN);
#define lcd_rw_high()   LCD_CTL_PORT |=  _BV(LCD_RW_PIN) 
#define lcd_rw_low()    LCD_CTL_PORT &= ~_BV(LCD_RW_PIN)
#define lcd_rs_high()   LCD_CTL_PORT |=  _BV(LCD_RS_PIN)
#define lcd_rs_low()    LCD_CTL_PORT &= ~_BV(LCD_RS_PIN)



#if LCD_LINES==1
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_1LINE 
#else
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_2LINES 
#endif



/* 
** function prototypes 
*/
static void lcd_e_toggle(void);

/*
** local functions
*/




/* toggle Enable Pin to initiate write */
static inline void lcd_e_toggle(void)
{
    lcd_e_high();
    lcd_e_delay();
    lcd_e_low();
}



/*************************************************************************
Low-level function to write byte to LCD controller
Input:    data   byte to write to LCD
          rs     1: write data    
                 0: write instruction
Returns:  none
*************************************************************************/

static void lcd_write(uint8_t data,uint8_t rs) 
{
    /* configure data pins as output */
  DDR(LCD_DATA_PORT)|=0xF<<LCD_DATA_SHIFT;

    if (rs)
        lcd_rs_high();                       /* RS=1: read data      */
    else
        lcd_rs_low();                        /* RS=0: read busy flag */
    lcd_rw_low();                           /* RW=0  write mode      */

       /* output high nibble first */
       LCD_DATA_PORT = ((data>>4)&0x0F)<<LCD_DATA_SHIFT;
       lcd_e_toggle();

       /* output low nibble */
       LCD_DATA_PORT = (data&0x0F)<<LCD_DATA_SHIFT;
       lcd_e_toggle();
       
    /* all data pins high (inactive) */
    LCD_DATA_PORT = 0x0F<<LCD_DATA_SHIFT;
}


/*************************************************************************
Low-level function to read byte from LCD controller
Input:    rs     1: read data    
                 0: read busy flag / address counter
Returns:  byte read from LCD controller
*************************************************************************/

static uint8_t lcd_read(uint8_t rs) 
{
    register uint8_t dataH, dataL;

    if (rs)
        lcd_rs_high();                       /* RS=1: read data      */
    else
        lcd_rs_low();                        /* RS=0: read busy flag */
    lcd_rw_high();                           /* RW=1  read mode      */
    
    /* configure data pins as input */
    DDR(LCD_DATA_PORT) &= ~(0xF<<LCD_DATA_SHIFT);  
    
    lcd_e_high();
    lcd_e_delay();
    dataH = PIN(LCD_DATA_PORT)>>LCD_DATA_SHIFT;              /* read high nibble first */
    lcd_e_low();
    
    lcd_e_delay();                           /* Enable 500ns low       */
    
    lcd_e_high();
    lcd_e_delay();
    dataL = PIN(LCD_DATA_PORT)>>LCD_DATA_SHIFT;              /* read low nibble        */
    lcd_e_low();
    
    return ( (dataH<<4) | (dataL&0x0F) );
}



/*************************************************************************
loops while lcd is busy, returns address counter
*************************************************************************/
static uint8_t lcd_waitbusy(void)

{
    register uint8_t c;
    
    /* wait until busy flag is cleared */
    while ( (c=lcd_read(0)) & (LCD_BUSY)) {}
    
    /* the address counter is updated 4us after the busy flag is cleared */
    delay(2);

    /* now read the address counter */
    return (lcd_read(0));  // return address counter
    
}/* lcd_waitbusy */


/*************************************************************************
Move cursor to the start of next line or to the first line if the cursor 
is already on the last line.
*************************************************************************/
static void lcd_newline(uint8_t pos)
{
    register uint8_t addressCounter;


#if LCD_LINES==1
    addressCounter = 0;
#endif
#if LCD_LINES==2
    if ( pos < (LCD_START_LINE2) )
        addressCounter = LCD_START_LINE2;
    else
        addressCounter = LCD_START_LINE1;
#endif
#if LCD_LINES==4
    if ( pos < LCD_START_LINE3 )
        addressCounter = LCD_START_LINE2;
    else if ( (pos >= LCD_START_LINE2) && (pos < LCD_START_LINE4) )
        addressCounter = LCD_START_LINE3;
    else if ( (pos >= LCD_START_LINE3) && (pos < LCD_START_LINE2) )
        addressCounter = LCD_START_LINE4;
    else 
        addressCounter = LCD_START_LINE1;
#endif
    lcd_command(LCD_DDRAM+addressCounter);

}/* lcd_newline */


/*
** PUBLIC FUNCTIONS 
*/

/*************************************************************************
Send LCD controller instruction command
Input:   instruction to send to LCD controller, see HD44780 data sheet
Returns: none
*************************************************************************/
void lcd_command(uint8_t cmd)
{
    lcd_waitbusy();
    lcd_write(cmd,0);
}


/*************************************************************************
Set cursor to specified position
Input:    x  horizontal position  (0: left most position)
          y  vertical position    (0: first line)
Returns:  none
*************************************************************************/
void lcd_gotoxy(uint8_t x, uint8_t y)
{
#if LCD_LINES==1
    lcd_command(LCD_DDRAM+LCD_START_LINE1+x);
#endif
#if LCD_LINES==2
    if ( y==0 ) 
        lcd_command(LCD_DDRAM+LCD_START_LINE1+x);
    else
        lcd_command(LCD_DDRAM+LCD_START_LINE2+x);
#endif
#if LCD_LINES==4
    if ( y==0 )
        lcd_command(LCD_DDRAM+LCD_START_LINE1+x);
    else if ( y==1)
        lcd_command(LCD_DDRAM+LCD_START_LINE2+x);
    else if ( y==2)
        lcd_command(LCD_DDRAM+LCD_START_LINE3+x);
    else /* y==3 */
        lcd_command(LCD_DDRAM+LCD_START_LINE4+x);
#endif

}/* lcd_gotoxy */


/*************************************************************************
*************************************************************************/
int lcd_getxy(void)
{
    return lcd_waitbusy();
}


/*************************************************************************
Clear display and set cursor to home position
*************************************************************************/
void lcd_clear(void)
{
  lcd_command(LCD_CLR);
}


/*************************************************************************
Set cursor to home position
*************************************************************************/
void lcd_home(void)
{
  lcd_command(LCD_HOME);
}


/*************************************************************************
Display character at current cursor position 
Input:    character to be displayed                                       
Returns:  none
*************************************************************************/
void lcd_putc(char c)
{
    uint8_t pos;


    pos = lcd_waitbusy();   // read busy-flag and address counter
    if (c=='\n')
    {
        lcd_newline(pos);
    }
    else
    {
#if LCD_WRAP_LINES==1
#if LCD_LINES==1
        if ( pos == LCD_START_LINE1+LCD_DISP_LENGTH )
            lcd_write(LCD_DDRAM+LCD_START_LINE1,0);
#elif LCD_LINES==2
        if ( pos == LCD_START_LINE1+LCD_DISP_LENGTH )
            lcd_write(LCD_DDRAM+LCD_START_LINE2,0);    
        else if ( pos == LCD_START_LINE2+LCD_DISP_LENGTH )
            lcd_write(LCD_DDRAM+LCD_START_LINE1,0);
#elif LCD_LINES==4
        if ( pos == LCD_START_LINE1+LCD_DISP_LENGTH )
            lcd_write(LCD_DDRAM+LCD_START_LINE2,0);    
        else if ( pos == LCD_START_LINE2+LCD_DISP_LENGTH )
            lcd_write(LCD_DDRAM+LCD_START_LINE3,0);
        else if ( pos == LCD_START_LINE3+LCD_DISP_LENGTH )
            lcd_write(LCD_DDRAM+LCD_START_LINE4,0);
        else if ( pos == LCD_START_LINE4+LCD_DISP_LENGTH )
            lcd_write(LCD_DDRAM+LCD_START_LINE1,0);
#endif
        lcd_waitbusy();
#endif
        lcd_write(c, 1);
    }

}/* lcd_putc */


/*************************************************************************
Display string without auto linefeed 
Input:    string to be displayed
Returns:  none
*************************************************************************/
void lcd_puts(const char *s)
/* print string on lcd (no auto linefeed) */
{
    register char c;

    while ( (c = *s++) ) {
        lcd_putc(c);
    }

}/* lcd_puts */


/*************************************************************************
Display string from program memory without auto linefeed 
Input:     string from program memory be be displayed                                        
Returns:   none
*************************************************************************/
void lcd_puts_p(const char *progmem_s)
/* print string from program memory on lcd (no auto linefeed) */
{
    register char c;

    while ( (c = PRG_RDB(progmem_s++)) ) {
        lcd_putc(c);
    }

}/* lcd_puts_p */


/*************************************************************************
Initialize display and select type of cursor 
Input:    dispAttr LCD_DISP_OFF            display off
                   LCD_DISP_ON             display on, cursor off
                   LCD_DISP_ON_CURSOR      display on, cursor on
                   LCD_DISP_CURSOR_BLINK   display on, cursor on flashing
Returns:  none
*************************************************************************/
void lcd_init(uint8_t dispAttr)
{
    /*
     *  Initialize LCD to 4 bit I/O mode
     */
         
    DDR(LCD_DATA_PORT) |= 0xF<<LCD_DATA_SHIFT;          /* all port bits as output */    
    DDR(LCD_CTL_PORT)|=_BV(LCD_RS_PIN)|_BV(LCD_E_PIN)|_BV(LCD_RW_PIN);  // assumes E, RS, and RW are on same port FIX

    delay(16000);        /* wait 16ms or more after power-on       */
    
    /* initial write to lcd is 8bit */
    LCD_DATA_PORT= (LCD_FUNCTION_8BIT_1LINE>>4)<<LCD_DATA_SHIFT;
    
    lcd_e_toggle();
    delay(4992);         /* delay, busy flag can't be checked here */
    
    LCD_DATA_PORT = (LCD_FUNCTION_8BIT_1LINE>>4)<<LCD_DATA_SHIFT;
    lcd_e_toggle();
    delay(64);           /* delay, busy flag can't be checked here */
    
    LCD_DATA_PORT = (LCD_FUNCTION_8BIT_1LINE>>4)<<LCD_DATA_SHIFT;
    lcd_e_toggle();
    delay(64);           /* delay, busy flag can't be checked here */
    
    LCD_DATA_PORT = (LCD_FUNCTION_4BIT_1LINE>>4)<<LCD_DATA_SHIFT; /* set IO mode to 4bit */
    lcd_e_toggle();
    delay(64);           /* some displays need this additional delay */
    


    /* from now the LCD only accepts 4 bit I/O, we can use lcd_command() */    
    lcd_command(LCD_FUNCTION_DEFAULT);      /* function set: display lines  */
    lcd_command(LCD_DISP_OFF);              /* display off                  */
    lcd_clear();                           /* display clear                */ 
    lcd_command(LCD_MODE_DEFAULT);          /* set entry mode               */
    lcd_command(dispAttr);                  /* display/cursor control       */

}/* lcd_init */

static int LcdWrite(char c)
{
  switch (c)
    {
    case '\014':
      lcd_clear();
      break;
      // could add more things here like BS, CR, etc.
    default:
      lcd_putc(c);
    }
  return 0;
}

void lcd_setstdio(void)
{
  fdevopen(LcdWrite,NULL,0);
}
