/* APP4 LCD library. Borrows ideas from Peter Fleury's LCD library
   which was based on Volker Oth's library.
*/
#ifndef APP4LCD_H
#define APP4LCD_H

#if (__GNUC__ * 100 + __GNUC_MINOR__) < 303
#error "This library requires AVR-GCC 3.3 or later, update to newer AVR-GCC compiler !"
#endif

#include <inttypes.h>
#include <avr/pgmspace.h>


/* 
 *  @name  Definitions for Display Size 
 *  Change these definitions to adapt setting to your display
 * If you use a 1 line display that is "split" set LCD_LINES=2
 * LCD_DISP_LENGTH=8 & LCD_WRAP_LINES=1
*/
#define LCD_LINES           2     /*< number of visible lines of the display */

#define LCD_DISP_LENGTH    8     /*< visibles characters per line of the display */
#define LCD_LINE_LENGTH  0x40     /*< internal line length of the display    */
#define LCD_START_LINE1  0x00     /*< DDRAM address of first char of line 1 */
#define LCD_START_LINE2  0x40     /*< DDRAM address of first char of line 2 */
#define LCD_START_LINE3  0x14     /*< DDRAM address of first char of line 3 */
#define LCD_START_LINE4  0x54     /*< DDRAM address of first char of line 4 */
#define LCD_WRAP_LINES      1     /*< 0: no wrap, 1: wrap at end of visibile line */


/* As set up, the LCD is connected like this:
   DB7 - PD.7
   DB6 - PD.6
   DB5 - PD.5
   DB4 - PD.4
   RS -  PB.2
   E  -  PB.1
   WR -  PB.0
*/
#define LCD_DATA_PORT    PORTD  /*< port for 4bit data */
#define LCD_DATA_SHIFT   4  /* Set to 4 for bit 7-4 or 0 for 3-0 */
#define LCD_CTL_PORT      PORTB  /*< port for control lines */
#define LCD_RS_PIN       2         /*< pin  for RS line */
#define LCD_RW_PIN       0         /*< pin  for RW line */
#define LCD_E_PIN        1         /*< pin  for Enable line */

/* instruction register bit positions, see HD44780U data sheet */
#define LCD_CLR               1
#define LCD_HOME              2
#define LCD_ENTRY_MODE        4
#define LCD_ENTRY_INC         2
#define LCD_ENTRY_SHIFT       1
#define LCD_ON                8 
#define LCD_ON_DISPLAY        4
#define LCD_ON_CURSOR         2
#define LCD_ON_BLINK          1
#define LCD_MOVE              16
#define LCD_MOVE_DISP         8
#define LCD_MOVE_RIGHT        4
#define LCD_FUNCTION          32
#define LCD_FUNCTION_8BIT     16
#define LCD_FUNCTION_2LINES   8
#define LCD_FUNCTION_10DOTS   4
#define LCD_CGRAM             64
#define LCD_DDRAM             128
#define LCD_BUSY              128

/* set entry mode: display shift on/off, dec/inc cursor move direction */
#define LCD_ENTRY_DEC            0x04   /* display shift off, dec cursor move dir */
#define LCD_ENTRY_DEC_SHIFT      0x05   /* display shift on,  dec cursor move dir */
#define LCD_ENTRY_INC_           0x06   /* display shift off, inc cursor move dir */
#define LCD_ENTRY_INC_SHIFT      0x07   /* display shift on,  inc cursor move dir */

/* display on/off, cursor on/off, blinking char at cursor position */
#define LCD_DISP_OFF             0x08   /* display off                            */
#define LCD_DISP_ON              0x0C   /* display on, cursor off                 */
#define LCD_DISP_ON_BLINK        0x0D   /* display on, cursor off, blink char     */
#define LCD_DISP_ON_CURSOR       0x0E   /* display on, cursor on                  */
#define LCD_DISP_ON_CURSOR_BLINK 0x0F   /* display on, cursor on, blink char      */

/* move cursor/shift display */
#define LCD_MOVE_CURSOR_LEFT     0x10   /* move cursor left  (decrement)          */
#define LCD_MOVE_CURSOR_RIGHT    0x14   /* move cursor right (increment)          */
#define LCD_MOVE_DISP_LEFT       0x18   /* shift display left                     */
#define LCD_MOVE_DISP_RIGHT      0x1C   /* shift display right                    */

/* function set: set interface data length and number of display lines */
#define LCD_FUNCTION_4BIT_1LINE  0x20   /* 4-bit interface, single line, 5x7 dots */
#define LCD_FUNCTION_4BIT_2LINES 0x28   /* 4-bit interface, dual line,   5x7 dots */
#define LCD_FUNCTION_8BIT_1LINE  0x30   /* 8-bit interface, single line, 5x7 dots */
#define LCD_FUNCTION_8BIT_2LINES 0x38   /* 8-bit interface, dual line,   5x7 dots */


#define LCD_MODE_DEFAULT     ((1<<LCD_ENTRY_MODE) | (1<<LCD_ENTRY_INC) )



/* Init LCD */
extern void lcd_init(uint8_t dispAttr);

/* Clear screen */
extern void lcd_clear(void);


/* Set cursor home */
extern void lcd_home(void);


/* Set x/y position */
extern void lcd_gotoxy(uint8_t x, uint8_t y);


/* Write character */
extern void lcd_putc(char c);


/* Put string */
extern void lcd_puts(const char *s);


/* Put string from program memory (no linefeed) */
extern void lcd_puts_p(const char *progmem_s);


/* Send LCD command (mostly internal) */
extern void lcd_command(uint8_t cmd);


/*
 macros for automatically storing string constant in program memory
*/
#define lcd_puts_P(__s)         lcd_puts_p(PSTR(__s))

/* Set LCD to be stdout (leaves stdin not set unfortunately) */
void lcd_setstdio(void);

#endif 
